package com.sun.java.util.jar.pack;

import java.util.Iterator;
import java.util.List;
import java.io.IOException;
import java.io.BufferedOutputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;

class ClassWriter
{
    int verbose;
    Package pkg;
    Package.Class cls;
    DataOutputStream out;
    ConstantPool.Index cpIndex;
    ConstantPool.Index bsmIndex;
    ByteArrayOutputStream buf;
    DataOutputStream bufOut;
    
    ClassWriter(Package.Class cls, OutputStream out) throws IOException {
        this.buf = new ByteArrayOutputStream();
        this.bufOut = new DataOutputStream(this.buf);
        this.pkg = cls.getPackage();
        this.cls = cls;
        this.verbose = this.pkg.verbose;
        this.out = new DataOutputStream(new BufferedOutputStream(out));
        this.cpIndex = ConstantPool.makeIndex(cls.toString(), cls.getCPMap());
        this.cpIndex.flattenSigs = true;
        if (cls.hasBootstrapMethods()) {
            this.bsmIndex = ConstantPool.makeIndex(this.cpIndex.debugName + ".BootstrapMethods", cls.getBootstrapMethodMap());
        }
        if (this.verbose > 1) {
            Utils.log.fine("local CP=" + ((this.verbose > 2) ? this.cpIndex.dumpString() : this.cpIndex.toString()));
        }
    }
    
    private void writeShort(int v) throws IOException {
        this.out.writeShort(v);
    }
    
    private void writeInt(int v) throws IOException {
        this.out.writeInt(v);
    }
    
    private void writeRef(ConstantPool.Entry entry) throws IOException {
        this.writeRef(entry, this.cpIndex);
    }
    
    private void writeRef(ConstantPool.Entry entry, ConstantPool.Index index) throws IOException {
        this.writeShort((entry == null) ? 0 : index.indexOf(entry));
    }
    
    void write() throws IOException {
        boolean n = false;
        try {
            if (this.verbose > 1) {
                Utils.log.fine("...writing " + this.cls);
            }
            this.writeMagicNumbers();
            this.writeConstantPool();
            this.writeHeader();
            this.writeMembers(false);
            this.writeMembers(true);
            this.writeAttributes(0, this.cls);
            this.out.flush();
            n = true;
        }
        finally {
            if (n == 0) {
                Utils.log.warning("Error on output of " + this.cls);
            }
        }
    }
    
    void writeMagicNumbers() throws IOException {
        this.writeInt(this.cls.magic);
        this.writeShort((int)this.cls.version.minor);
        this.writeShort((int)this.cls.version.major);
    }
    
    void writeConstantPool() throws IOException {
        ConstantPool.Entry[] cpMap = this.cls.cpMap;
        this.writeShort(cpMap.length);
        int i = 0;
        while (i < cpMap.length) {
            ConstantPool.Entry obj = cpMap[i];
            assert obj == null == (i == 0 || (cpMap[i - 1] != null && cpMap[i - 1].isDoubleWord()));
            if (obj != null) {
                int tag = obj.getTag();
                if (this.verbose > 2) {
                    Utils.log.fine("   CP[" + i + "] = " + obj);
                }
                this.out.write(tag);
                switch (tag) {
                    case 13: {
                        throw new AssertionError("CP should have Signatures remapped to Utf8");
                    }
                    case 1: {
                        this.out.writeUTF(obj.stringValue());
                        break;
                    }
                    case 3: {
                        this.out.writeInt(((ConstantPool.NumberEntry)obj).numberValue().intValue());
                        break;
                    }
                    case 4: {
                        this.out.writeInt(Float.floatToRawIntBits(((ConstantPool.NumberEntry)obj).numberValue().floatValue()));
                        break;
                    }
                    case 5: {
                        this.out.writeLong(((ConstantPool.NumberEntry)obj).numberValue().longValue());
                        break;
                    }
                    case 6: {
                        this.out.writeLong(Double.doubleToRawLongBits(((ConstantPool.NumberEntry)obj).numberValue().doubleValue()));
                        break;
                    }
                    case 7:
                    case 8:
                    case 16: {
                        this.writeRef(obj.getRef(0));
                        break;
                    }
                    case 15: {
                        ConstantPool.MethodHandleEntry methodHandleEntry = (ConstantPool.MethodHandleEntry)obj;
                        this.out.writeByte(methodHandleEntry.refKind);
                        this.writeRef(methodHandleEntry.getRef(0));
                        break;
                    }
                    case 9:
                    case 10:
                    case 11:
                    case 12: {
                        this.writeRef(obj.getRef(0));
                        this.writeRef(obj.getRef(1));
                        break;
                    }
                    case 18: {
                        this.writeRef(obj.getRef(0), this.bsmIndex);
                        this.writeRef(obj.getRef(1));
                        break;
                    }
                    case 17: {
                        throw new AssertionError("CP should have BootstrapMethods moved to side-table");
                    }
                    default: {
                        throw new IOException("Bad constant pool tag " + tag);
                    }
                }
            }
            ++i;
        }
    }
    
    void writeHeader() throws IOException {
        this.writeShort(this.cls.flags);
        this.writeRef(this.cls.thisClass);
        this.writeRef(this.cls.superClass);
        this.writeShort(this.cls.interfaces.length);
        for (int i = 0; i < this.cls.interfaces.length; ++i) {
            this.writeRef(this.cls.interfaces[i]);
        }
    }
    
    void writeMembers(boolean b) throws IOException {
        List list;
        if (!b) {
            list = this.cls.getFields();
        }
        else {
            list = this.cls.getMethods();
        }
        this.writeShort(list.size());
        Iterator<Object> iterator = list.iterator();
        while (iterator.hasNext()) {
            this.writeMember((Package.Class.Member)iterator.next(), b);
        }
    }
    
    void writeMember(Package.Class.Member obj, boolean b) throws IOException {
        if (this.verbose > 2) {
            Utils.log.fine("writeMember " + obj);
        }
        this.writeShort(obj.flags);
        this.writeRef(obj.getDescriptor().nameRef);
        this.writeRef(obj.getDescriptor().typeRef);
        this.writeAttributes(b ? 2 : 1, obj);
    }
    
    private void reorderBSMandICS(Attribute.Holder holder) {
        Attribute attribute = holder.getAttribute(Package.attrBootstrapMethodsEmpty);
        if (attribute == null) {
            return;
        }
        Attribute attribute2 = holder.getAttribute(Package.attrInnerClassesEmpty);
        if (attribute2 == null) {
            return;
        }
        int index = holder.attributes.indexOf(attribute);
        int index2 = holder.attributes.indexOf(attribute2);
        if (index > index2) {
            holder.attributes.remove(attribute);
            holder.attributes.add(index2, attribute);
        }
        return;
    }
    
    void writeAttributes(int n, Attribute.Holder holder) throws IOException {
        if (holder.attributes == null) {
            this.writeShort(0);
            return;
        }
        if ((holder instanceof Package.Class)) {
            this.reorderBSMandICS(holder);
        }
        this.writeShort(holder.attributes.size());
        for (Attribute attribute : holder.attributes) {
            attribute.finishRefs(this.cpIndex);
            this.writeRef(attribute.getNameRef());
            if (attribute.layout() == Package.attrCodeEmpty || attribute.layout() == Package.attrBootstrapMethodsEmpty || attribute.layout() == Package.attrInnerClassesEmpty) {
                DataOutputStream out = this.out;
                assert this.out != this.bufOut;
                this.buf.reset();
                this.out = this.bufOut;
                if ("Code".equals(attribute.name())) {
                    this.writeCode(((Package.Class.Method)holder).code);
                }
                else if ("BootstrapMethods".equals(attribute.name())) {
                    assert holder == this.cls;
                    this.writeBootstrapMethods(this.cls);
                }
                else if ("InnerClasses".equals(attribute.name())) {
                    assert holder == this.cls;
                    this.writeInnerClasses(this.cls);
                }
                else {
                    throw new AssertionError();
                }
                this.out = out;
                if (this.verbose > 2) {
                    Utils.log.fine("Attribute " + attribute.name() + " [" + this.buf.size() + "]");
                }
                this.writeInt(this.buf.size());
                this.buf.writeTo(this.out);
            }
            else {
                if (this.verbose > 2) {
                    Utils.log.fine("Attribute " + attribute.name() + " [" + attribute.size() + "]");
                }
                this.writeInt(attribute.size());
                this.out.write(attribute.bytes());
            }
        }
        return;
    }
    
    void writeCode(Code code) throws IOException {
        code.finishRefs(this.cpIndex);
        this.writeShort(code.max_stack);
        this.writeShort(code.max_locals);
        this.writeInt(code.bytes.length);
        this.out.write(code.bytes);
        int handlerCount = code.getHandlerCount();
        this.writeShort(handlerCount);
        for (int i = 0; i < handlerCount; ++i) {
            this.writeShort(code.handler_start[i]);
            this.writeShort(code.handler_end[i]);
            this.writeShort(code.handler_catch[i]);
            this.writeRef(code.handler_class[i]);
        }
        this.writeAttributes(3, code);
    }
    
    void writeBootstrapMethods(Package.Class class1) throws IOException {
        List<ConstantPool.BootstrapMethodEntry> bootstrapMethods = class1.getBootstrapMethods();
        this.writeShort(bootstrapMethods.size());
        for (ConstantPool.BootstrapMethodEntry bootstrapMethodEntry : bootstrapMethods) {
            this.writeRef(bootstrapMethodEntry.bsmRef);
            this.writeShort(bootstrapMethodEntry.argRefs.length);
            ConstantPool.Entry[] argRefs = bootstrapMethodEntry.argRefs;
            for (int length = argRefs.length, i = 0; i < length; ++i) {
                this.writeRef(argRefs[i]);
            }
        }
    }
    
    void writeInnerClasses(Package.Class class1) throws IOException {
        List<Package.InnerClass> innerClasses = class1.getInnerClasses();
        this.writeShort(innerClasses.size());
        for (Package.InnerClass innerClass : innerClasses) {
            this.writeRef(innerClass.thisClass);
            this.writeRef(innerClass.outerClass);
            this.writeRef(innerClass.name);
            this.writeShort(innerClass.flags);
        }
    }
}
